package org.openspaces.cassandraeds;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.gigaspaces.datasource.BulkDataPersister;
import com.gigaspaces.datasource.BulkItem;
import com.gigaspaces.datasource.DataIterator;
import com.gigaspaces.datasource.DataSourceException;
import com.gigaspaces.datasource.ManagedDataSource;
import com.gigaspaces.datasource.SQLDataProvider;
import com.gigaspaces.metadata.SpaceTypeDescriptor;
import com.j_spaces.core.IGSEntry;
import com.j_spaces.core.client.SQLQuery;
import com.jolbox.bonecp.BoneCPDataSource;

/**
 * Cassandra EDS.  Constructors define usage.  2 arg constructor for mirror eds.  3 arg constructors
 * for space EDS.  Which one to use depends on whether using SpaceDocuments or POJOs.
 * 
 * @author DeWayne
 *
 */
public class CassandraEDS implements ManagedDataSource<Object>,BulkDataPersister,SQLDataProvider<Object>{
	private static final Logger logger = Logger.getLogger(CassandraEDS.class.getName());
	private final BoneCPDataSource connectionPool;
	private final FieldSerializer fieldSerializer;
	private final List<Class<?>> classes = new ArrayList<Class<?>>();
	private final SpaceTypeDescriptor[] doctypes;

	private final ConcurrentHashMap<String, String> insertSQLCache = new ConcurrentHashMap<String, String> ();
//	private final ConcurrentHashMap<String, String> updateSQLCache = new ConcurrentHashMap<String, String> ();
	
	/**
	 * Constructor for mirror EDS
	 * 
	 */
	public CassandraEDS( BoneCPDataSource connectionPool, FieldSerializer fieldSerializer){
		this(connectionPool, fieldSerializer, (SpaceTypeDescriptor[])null);
	}
			

	/**
	 * Constructor for POJO based space EDSs
	 * 
	 */
	public CassandraEDS( BoneCPDataSource connectionPool, FieldSerializer fieldSerializer, String[] classes) throws Exception{
		this(connectionPool, fieldSerializer, (SpaceTypeDescriptor[])null);
		for(String clazz:classes){
			this.classes.add(Class.forName(clazz));
		}
	}
	
	/**
	 * Constructor for SpaceDocument based space EDSs
	 * 
	 */
	public CassandraEDS( BoneCPDataSource connectionPool, FieldSerializer fieldSerializer,
			SpaceTypeDescriptor[] doctypes){
		this.connectionPool = connectionPool;
		this.fieldSerializer = fieldSerializer;
		this.doctypes = doctypes;
	}

	/**
	 * Checks to see if Cassandra is listening.
	 */
	public void init(Properties properties) throws DataSourceException {
		try{
			Connection test = connectionPool.getConnection();
			if(test==null){
				throw new DataSourceException("no connections available");
			}
			test.close();
		}
		catch(Exception e){
			throw new DataSourceException(e);
		}
	}
	
	@Override
	public void shutdown() throws DataSourceException {
		connectionPool.close();
	}

	@Override
	public void executeBulk(List<BulkItem> bulkItems) throws DataSourceException {
		Connection con = null;
		try {
			con = connectionPool.getConnection();
			for (BulkItem bulkItem : bulkItems) {
				IGSEntry item = (IGSEntry) bulkItem.getItem();
				String clazzName = item.getClassName().replaceAll("\\.", "_");
				
				String ID = item.getFieldValue(item.getPrimaryKeyName()).toString();
				
				if(item.getFieldType(item.getPrimaryKeyName()).equals("java.lang.String")){
					ID= '\'' + ID + '\'';
				}

				switch (bulkItem.getOperation()) {
				case BulkItem.REMOVE:
					String deleteQL = "DELETE FROM " + clazzName + " WHERE KEY = " + ID;
					try {
						if(logger.isLoggable(Level.FINE))
							logger.fine("removing :  " + deleteQL);
						
						executeCQL(con, deleteQL);
					} catch (Exception e) {
						logger.log(Level.SEVERE, e.getMessage(), e);
					}
					break;
				case BulkItem.WRITE:
					int fldCount = item.getFieldsNames().length;
					StringBuilder insertQLBuf = new StringBuilder();
					final String insertQL = insertSQLCache.get(clazzName);
					if (insertQL == null)
					{
						insertQLBuf.append( "INSERT INTO " ).append( clazzName).append(" (KEY,");
						for (int i = 0; i < fldCount ; i++) {
							insertQLBuf.append( " '").append( item.getFieldsNames()[i]).append( "',");
						}
						insertQLBuf.deleteCharAt(insertQLBuf.length()-1);
						insertQLBuf.append( ") VALUES (");
						insertSQLCache.put(clazzName, insertQLBuf.toString());
					}
					else{
						insertQLBuf.append(insertQL);
					}

					insertQLBuf.append(ID).append(',');
					for (int i = 0; i < fldCount; i++) {
						insertQLBuf.append(getValue(item.getFieldsNames()[i],item.getFieldValue(i))).append(',');
					}
					insertQLBuf.deleteCharAt(insertQLBuf.length()-1).append(')').toString();
					try {
						if(logger.isLoggable(Level.FINE))
							logger.fine("inserting: " + insertQLBuf);
						
						executeCQL(con, insertQLBuf.toString());
					} catch (Exception e) {
						logger.log(Level.SEVERE, e.getMessage(), e);
					}
					break;

				case BulkItem.UPDATE:
					fldCount = item.getFieldsNames().length;
					StringBuilder updateQL=new StringBuilder("UPDATE ");
					updateQL.append(clazzName).append(" SET");
					for (int i = 0; i < fldCount ; i++) {
						updateQL.append(" '")
						.append(item.getFieldsNames()[i])
						.append("'=")
						.append(getValue(item.getFieldsNames()[i],item.getFieldsValues()[i]))
						.append(',');
					}
					updateQL.deleteCharAt(updateQL.length()-1);

					updateQL.append(" WHERE KEY=").append(ID);
					try {
						if(logger.isLoggable(Level.FINE))
							logger.fine("updating: " + updateQL);
						
						executeCQL(con, updateQL.toString());
					} catch (Exception e) {
						logger.log(Level.SEVERE, e.getMessage(), e);
					}
					break;
				}
			}
		} catch (SQLException e) {
			logger.log(Level.SEVERE, e.getMessage(), e);
		}
		finally{
			try {
				if(con != null)
					con.close();
			} catch (SQLException e) {
				logger.log(Level.WARNING,e.getMessage(),e);
			}
		}
	}

	@Override
	public DataIterator<Object> initialLoad() throws DataSourceException {
		Connection con=null;
		try {
			con = connectionPool.getConnection();

			if (doctypes == null) 
				return DataIterators.newMultiClassIterator(classes,fieldSerializer,con);
			
			return DataIterators.newMultiDocumentIterator(doctypes,fieldSerializer,con);
		} catch (SQLException e) {
			throw new DataSourceException(e);
		}
		finally{
			try {
				if(con != null)
					con.close();
			} catch (SQLException e) {
				logger.log(Level.WARNING,e.getMessage(),e);
			}
		}
	}

	@Override
	public DataIterator<Object> iterator(SQLQuery<Object> query)throws DataSourceException {
		Connection con = null;
		try{
			con = connectionPool.getConnection();

			if (doctypes == null)
				return DataIterators.newSQLIterator(query, fieldSerializer, classes.toArray(new Class<?>[classes.size()]), con); 

			return DataIterators.newSQLDocumentIterator(query,fieldSerializer,doctypes,con);
		} catch (SQLException e) {
			throw new DataSourceException(e);
		}
		finally{
			try {
				if(con != null)
					con.close();
			} catch (SQLException e) {
				logger.log(Level.WARNING,e.getMessage(),e);
			}
		}
	}


	/**
	 * Gets the string value of a field.  For non-java types,
	 * uses supplied serializer.  Presumes only a single level
	 * of object nesting.
	 * 
	 * @param fieldName the field to get the value of
	 * @param val the object value of the field to be converted to string
	 * @return a java.lang.String value of the field
	 */
	private String getValue(String fieldName, Object val)
	{
		if (val == null)
			return "''";

		//Note - presumes non-collection fields
		if(logger.isLoggable(Level.FINE))
			logger.fine("getValue val type:" + val.getClass().getName());
		
		if(val.getClass().getName().startsWith("java.lang.")){
			String str = val.toString();
			if (str.indexOf('\'') > 0)
				return '\'' + str.replaceAll("'", "''") + '\'' ;

			return '\'' + str + '\'';
		}
		return '\'' + fieldSerializer.serialize(val) + '\'';
	}

	private void executeCQL(Connection con, String cql) throws SQLException {
		Statement statement = con.createStatement();
		statement.execute(cql);
		statement.close();
	}
}